@@ -4,7 +4,7 @@
  * Copyright (C) 2005-2007 Mikael Berthe <mikael@lilotux.net>
  * Some parts initially came from the centericq project:
  * Copyright (C) 2002-2005 by Konstantin Klyagin <konst@konst.org.ua>
- * Some small parts come from the Gaim project <http://gaim.sourceforge.net/>
+ * Some small parts come from the Pidgin project <http://pidgin.im/>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -33,6 +33,7 @@
 #include "settings.h"
 #include "hbuf.h"
 #include "commands.h"
+#include "hooks.h"
 
 #ifdef ENABLE_HGCSET
 # include "hgcset.h"
@@ -46,6 +47,8 @@ xmlnode rosternotes;
 
 static GSList *iqs_list;
 
+time_t iqlast; // last message/status change time
+
 // Enum for vCard attributes
 enum vcard_attr {
   vcard_home    = 1<<0,
@@ -110,14 +113,15 @@ const struct adhoc_status adhoc_status_list[] = {
 const char *entity_version(void)
 {
   static char *ver;
+  const char *PVERSION = PACKAGE_VERSION; // "+xxx";
 
   if (ver)
     return ver;
 
 #ifdef HGCSET
-  ver = g_strdup_printf("%s-%s", PACKAGE_VERSION, HGCSET);
+  ver = g_strdup_printf("%s-%s", PVERSION, HGCSET);
 #else
-  ver = g_strdup(PACKAGE_VERSION);
+  ver = g_strdup(PVERSION);
 #endif
 
   return ver;
@@ -243,13 +247,6 @@ void jb_iqs_display_list(void)
   scr_LogPrint(LPRINT_LOGNORM, "End of IQ list.");
 }
 
-static void request_roster(void)
-{
-  eviqs *iqn = iqs_new(JPACKET__GET, NS_ROSTER, "Roster", IQS_DEFAULT_TIMEOUT);
-  jab_send(jc, iqn->xmldata);
-  iqs_del(iqn->id); // XXX
-}
-
 static void handle_iq_roster(xmlnode x)
 {
   xmlnode y;
@@ -260,6 +257,7 @@ static void handle_iq_roster(xmlnode x)
   guint roster_type;
 
   for (y = xmlnode_get_tag(x, "item"); y; y = xmlnode_get_nextsibling(y)) {
+    char *name_tmp = NULL;
 
     fjid = xmlnode_get_attrib(y, "jid");
     name = xmlnode_get_attrib(y, "name");
@@ -293,8 +291,16 @@ static void handle_iq_roster(xmlnode x)
     if (ask && !strcmp(ask, "subscribe"))
       esub |= sub_pending;
 
-    if (!name)
-      name = cleanalias;
+    if (!name) {
+      if (!settings_opt_get_int("roster_hide_domain")) {
+        name = cleanalias;
+      } else {
+        char *p;
+        name = name_tmp = g_strdup(cleanalias);
+        p = strchr(name_tmp, JID_DOMAIN_SEPARATOR);
+        if (p)  *p = '\0';
+      }
+    }
 
     // Tricky... :-\  My guess is that if there is no JID_DOMAIN_SEPARATOR,
     // this is an agent.
@@ -303,8 +309,9 @@ static void handle_iq_roster(xmlnode x)
     else
       roster_type = ROSTER_TYPE_AGENT;
 
-    roster_add_user(cleanalias, name, group, roster_type, esub);
+    roster_add_user(cleanalias, name, group, roster_type, esub, 1);
 
+    g_free(name_tmp);
     g_free(cleanalias);
   }
 
@@ -314,6 +321,42 @@ static void handle_iq_roster(xmlnode x)
     scr_UpdateBuddyWindow();
 }
 
+//  This callback is reached when mcabber receives the first roster update
+// after the connection.
+static int iqscallback_gotroster(eviqs *iqp, xmlnode xml_result, guint iqcontext)
+{
+  xmlnode x;
+  char *ns;
+
+  // Leave now if we cannot process xml_result
+  if (!xml_result || iqcontext) return -1;
+
+  // Only execute the hook if the roster has been successfully retrieved
+  if (iqcontext != IQS_CONTEXT_RESULT)
+    return 0;
+
+  x = xmlnode_get_tag(xml_result, "query");
+  if (!x)
+    return -1;
+
+  ns = xmlnode_get_attrib(x, "xmlns");
+  if (ns && !strcmp(ns, NS_ROSTER))
+    handle_iq_roster(x);
+
+  // Post-login stuff
+  jb_setprevstatus();
+  hook_execute_internal("hook-post-connect");
+
+  return 0;
+}
+
+static void request_roster(void)
+{
+  eviqs *iqn = iqs_new(JPACKET__GET, NS_ROSTER, "Roster", IQS_DEFAULT_TIMEOUT);
+  iqn->callback = &iqscallback_gotroster;
+  jab_send(jc, iqn->xmldata);
+}
+
 static int iqscallback_version(eviqs *iqp, xmlnode xml_result, guint iqcontext)
 {
   xmlnode ansqry;
@@ -344,26 +387,29 @@ static int iqscallback_version(eviqs *iqp, xmlnode xml_result, guint iqcontext)
   p = strchr(bjid, JID_RESOURCE_SEPARATOR);
   if (p) *p = '\0';
 
-  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO);
+  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO, 0);
   g_free(buf);
 
   // Get result data...
   p = xmlnode_get_tag_data(ansqry, "name");
   if (p) {
     buf = g_strdup_printf("Name:    %s", p);
-    scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_NONE);
+    scr_WriteIncomingMessage(bjid, buf,
+                             0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
     g_free(buf);
   }
   p = xmlnode_get_tag_data(ansqry, "version");
   if (p) {
     buf = g_strdup_printf("Version: %s", p);
-    scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_NONE);
+    scr_WriteIncomingMessage(bjid, buf,
+                             0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
     g_free(buf);
   }
   p = xmlnode_get_tag_data(ansqry, "os");
   if (p) {
     buf = g_strdup_printf("OS:      %s", p);
-    scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_NONE);
+    scr_WriteIncomingMessage(bjid, buf,
+                             0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
     g_free(buf);
   }
   return 0;
@@ -409,26 +455,29 @@ static int iqscallback_time(eviqs *iqp, xmlnode xml_result, guint iqcontext)
   p = strchr(bjid, JID_RESOURCE_SEPARATOR);
   if (p) *p = '\0';
 
-  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO);
+  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO, 0);
   g_free(buf);
 
   // Get result data...
   p = xmlnode_get_tag_data(ansqry, "utc");
   if (p) {
     buf = g_strdup_printf("UTC:  %s", p);
-    scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_NONE);
+    scr_WriteIncomingMessage(bjid, buf,
+                             0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
     g_free(buf);
   }
   p = xmlnode_get_tag_data(ansqry, "tz");
   if (p) {
     buf = g_strdup_printf("TZ:   %s", p);
-    scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_NONE);
+    scr_WriteIncomingMessage(bjid, buf,
+                             0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
     g_free(buf);
   }
   p = xmlnode_get_tag_data(ansqry, "display");
   if (p) {
     buf = g_strdup_printf("Time: %s", p);
-    scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_NONE);
+    scr_WriteIncomingMessage(bjid, buf,
+                             0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
     g_free(buf);
   }
   return 0;
@@ -474,7 +523,7 @@ static int iqscallback_last(eviqs *iqp, xmlnode xml_result, guint iqcontext)
   p = strchr(bjid, JID_RESOURCE_SEPARATOR);
   if (p) *p = '\0';
 
-  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO);
+  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO, 0);
   g_free(buf);
 
   // Get result data...
@@ -493,16 +542,17 @@ static int iqscallback_last(eviqs *iqp, xmlnode xml_result, guint iqcontext)
     g_string_append_printf(sbuf, "%02ld:", s/3600L);
     s %= 3600L;
     g_string_append_printf(sbuf, "%02ld:%02ld", s/60L, s%60L);
-    scr_WriteIncomingMessage(bjid, sbuf->str, 0, HBB_PREFIX_NONE);
+    scr_WriteIncomingMessage(bjid, sbuf->str,
+                             0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
     g_string_free(sbuf, TRUE);
   } else {
     scr_WriteIncomingMessage(bjid, "No idle time reported.",
-                             0, HBB_PREFIX_NONE);
+                             0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
   }
   p = xmlnode_get_data(ansqry);
   if (p) {
     buf = g_strdup_printf("Status message: %s", p);
-    scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO);
+    scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO, 0);
     g_free(buf);
   }
   return 0;
@@ -537,7 +587,7 @@ static void display_vcard_item(const char *bjid, const char *label,
                         (vcard_attrib & vcard_pref ? "[pref]" : ""),
                         (vcard_attrib ? " " : ""),
                         text);
-  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_NONE);
+  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO | HBB_PREFIX_CONT, 0);
   g_free(buf);
 }
 
@@ -654,6 +704,7 @@ static int iqscallback_vcard(eviqs *iqp, xmlnode xml_result, guint iqcontext)
   ansqry = xmlnode_get_tag(xml_result, "vCard");
   if (!ansqry) {
     scr_LogPrint(LPRINT_LOGNORM, "Empty IQ:vCard result!");
+    g_free(buf);
     return 0;
   }
 
@@ -661,7 +712,7 @@ static int iqscallback_vcard(eviqs *iqp, xmlnode xml_result, guint iqcontext)
   p = strchr(bjid, JID_RESOURCE_SEPARATOR);
   if (p) *p = '\0';
 
-  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO);
+  scr_WriteIncomingMessage(bjid, buf, 0, HBB_PREFIX_INFO, 0);
   g_free(buf);
 
   // Get result data...
@@ -672,27 +723,23 @@ static int iqscallback_vcard(eviqs *iqp, xmlnode xml_result, guint iqcontext)
 void request_vcard(const char *bjid)
 {
   eviqs *iqn;
-  char *barejid;
-
-  barejid = jidtodisp(bjid);
 
   // Create a new IQ structure.  We use NULL for the namespace because
   // we'll have to use a special tag, not the usual "query" one.
   iqn = iqs_new(JPACKET__GET, NULL, "vcard", IQS_DEFAULT_TIMEOUT);
-  xmlnode_put_attrib(iqn->xmldata, "to", barejid);
+  xmlnode_put_attrib(iqn->xmldata, "to", bjid);
   // Remove the useless <query/> tag, and insert a vCard one.
   xmlnode_hide(xmlnode_get_tag(iqn->xmldata, "query"));
   xmlnode_put_attrib(xmlnode_insert_tag(iqn->xmldata, "vCard"),
                      "xmlns", NS_VCARD);
   iqn->callback = &iqscallback_vcard;
   jab_send(jc, iqn->xmldata);
-
-  g_free(barejid);
 }
 
 static void storage_bookmarks_parse_conference(xmlnode xmldata)
 {
   const char *fjid, *name, *autojoin;
+  const char *pstatus, *awhois;
   char *bjid;
   GSList *room_elt;
 
@@ -701,13 +748,16 @@ static void storage_bookmarks_parse_conference(xmlnode xmldata)
     return;
   name = xmlnode_get_attrib(xmldata, "name");
   autojoin = xmlnode_get_attrib(xmldata, "autojoin");
+  awhois = xmlnode_get_attrib(xmldata, "autowhois");
+  pstatus = xmlnode_get_tag_data(xmldata, "print_status");
 
   bjid = jidtodisp(fjid); // Bare jid
 
   // Make sure this is a room (it can be a conversion user->room)
   room_elt = roster_find(bjid, jidsearch, 0);
   if (!room_elt) {
-    room_elt = roster_add_user(bjid, name, NULL, ROSTER_TYPE_ROOM, sub_none);
+    room_elt = roster_add_user(bjid, name, NULL, ROSTER_TYPE_ROOM,
+                               sub_none, -1);
   } else {
     buddy_settype(room_elt->data, ROSTER_TYPE_ROOM);
     /*
@@ -719,6 +769,25 @@ static void storage_bookmarks_parse_conference(xmlnode xmldata)
     */
   }
 
+  // Set the print_status and auto_whois values
+  if (pstatus) {
+    enum room_printstatus i;
+    for (i = status_none; i <= status_all; i++)
+      if (!strcasecmp(pstatus, strprintstatus[i]))
+        break;
+    if (i <= status_all)
+      buddy_setprintstatus(room_elt->data, i);
+  }
+  if (awhois) {
+    enum room_autowhois i = autowhois_default;
+    if (!strcmp(awhois, "1"))
+      i = autowhois_on;
+    else if (!strcmp(awhois, "0"))
+      i = autowhois_off;
+    if (i != autowhois_default)
+      buddy_setautowhois(room_elt->data, i);
+  }
+
   // Is autojoin set?
   // If it is, we'll look up for more information (nick? password?) and
   // try to join the room.
@@ -728,7 +797,7 @@ static void storage_bookmarks_parse_conference(xmlnode xmldata)
     nick = xmlnode_get_tag_data(xmldata, "nick");
     passwd = xmlnode_get_tag_data(xmldata, "password");
     if (!nick || !*nick)
-      nick = tmpnick = default_muc_nickname();
+      nick = tmpnick = default_muc_nickname(NULL);
     // Let's join now
     scr_LogPrint(LPRINT_LOGNORM, "Auto-join bookmark <%s>", bjid);
     jb_room_join(bjid, nick, passwd);
@@ -879,33 +948,14 @@ int iqscallback_auth(eviqs *iqp, xmlnode xml_result, guint iqcontext)
 
 static void handle_iq_result(jconn conn, char *from, xmlnode xmldata)
 {
-  xmlnode x;
-  char *id;
-  char *ns;
+  char *id = xmlnode_get_attrib(xmldata, "id");
 
-  id = xmlnode_get_attrib(xmldata, "id");
   if (!id) {
     scr_LogPrint(LPRINT_LOG, "IQ result stanza with no ID, ignored.");
     return;
   }
 
-  if (!iqs_callback(id, xmldata, IQS_CONTEXT_RESULT))
-    return;
-
-  x = xmlnode_get_tag(xmldata, "query");
-  if (!x) return;
-
-  ns = xmlnode_get_attrib(x, "xmlns");
-  if (!ns) return;
-
-  if (!strcmp(ns, NS_ROSTER)) {
-    handle_iq_roster(x);
-
-    // Post-login stuff
-    // Usually we request the roster only at connection time
-    // so we should be there only once.  (That's ugly, however)
-    jb_setprevstatus();
-  }
+  (void)iqs_callback(id, xmldata, IQS_CONTEXT_RESULT);
 }
 
 // FIXME  highly duplicated code
@@ -928,6 +978,26 @@ static void send_iq_not_implemented(jconn conn, char *from, xmlnode xmldata)
   xmlnode_free(x);
 }
 
+// FIXME  highly duplicated code
+static void send_iq_not_available(jconn conn, char *from, xmlnode xmldata)
+{
+  xmlnode x, y, z;
+  // Not available.
+  x = xmlnode_dup(xmldata);
+  xmlnode_put_attrib(x, "to", xmlnode_get_attrib(xmldata, "from"));
+  xmlnode_hide_attrib(x, "from");
+
+  xmlnode_put_attrib(x, "type", TMSG_ERROR);
+  y = xmlnode_insert_tag(x, TMSG_ERROR);
+  xmlnode_put_attrib(y, "code", "503");
+  xmlnode_put_attrib(y, "type", "cancel");
+  z = xmlnode_insert_tag(y, "service-unavailable");
+  xmlnode_put_attrib(z, "xmlns", NS_XMPP_STANZAS);
+
+  jab_send(conn, x);
+  xmlnode_free(x);
+}
+
 /*
 static void send_iq_commands_bad_action(jconn conn, char *from, xmlnode xmldata)
 {
@@ -1299,6 +1369,11 @@ static void disco_info_set_ext(xmlnode ansquery, const char *ext)
     xmlnode_put_attrib(xmlnode_insert_tag(ansquery, "feature"),
                        "var", NS_CHATSTATES);
   }
+  if (!strcasecmp(ext, "iql")) {
+    // I guess it's ok to send this even if it's not compiled in.
+    xmlnode_put_attrib(xmlnode_insert_tag(ansquery, "feature"),
+                       "var", NS_LAST);
+  }
 }
 
 //  disco_info_set_default(ansquery, entitycaps)
@@ -1333,12 +1408,17 @@ static void disco_info_set_default(xmlnode ansquery, guint entitycaps)
 #endif
   xmlnode_put_attrib(xmlnode_insert_tag(ansquery, "feature"),
                      "var", NS_TIME);
+  xmlnode_put_attrib(xmlnode_insert_tag(ansquery, "feature"),
+                     "var", NS_XMPP_TIME);
   xmlnode_put_attrib(xmlnode_insert_tag(ansquery, "feature"),
                      "var", NS_VERSION);
   xmlnode_put_attrib(xmlnode_insert_tag(ansquery, "feature"),
                      "var", NS_PING);
   xmlnode_put_attrib(xmlnode_insert_tag(ansquery, "feature"),
                      "var", NS_COMMANDS);
+  if (!entitycaps)
+    xmlnode_put_attrib(xmlnode_insert_tag(ansquery, "feature"),
+                       "var", NS_LAST);
 }
 
 static void handle_iq_disco_info(jconn conn, char *from, const char *id,
@@ -1369,6 +1449,33 @@ static void handle_iq_disco_info(jconn conn, char *from, const char *id,
   xmlnode_free(x);
 }
 
+inline double seconds_since_last_use(void)
+{
+  return difftime(time(NULL), iqlast);
+}
+
+static void handle_iq_last(jconn conn, char *from, const char *id,
+                           xmlnode xmldata)
+{
+  xmlnode x;
+  xmlnode myquery;
+  char *seconds;
+
+  scr_LogPrint(LPRINT_LOGNORM, "Received an IQ last time request from <%s>",
+               from);
+
+  x = jutil_iqnew(JPACKET__RESULT, NS_LAST);
+  xmlnode_put_attrib(x, "id", id);
+  xmlnode_put_attrib(x, "to", xmlnode_get_attrib(xmldata, "from"));
+  myquery = xmlnode_get_tag(x, "query");
+  seconds = g_strdup_printf("%.0f", seconds_since_last_use());
+  xmlnode_put_attrib(myquery, "seconds", seconds);
+  g_free(seconds);
+
+  jab_send(jc, x);
+  xmlnode_free(x);
+}
+
 static void handle_iq_ping(jconn conn, char *from, const char *id,
                            xmlnode xmldata)
 {
@@ -1412,9 +1519,9 @@ static void handle_iq_version(jconn conn, char *from, const char *id,
   xmlnode_free(x);
 }
 
-// This function borrows some code from the Gaim project
+// This function borrows some code from the Pidgin project
 static void handle_iq_time(jconn conn, char *from, const char *id,
-                              xmlnode xmldata)
+                           xmlnode xmldata)
 {
   xmlnode x;
   xmlnode myquery;
@@ -1457,7 +1564,66 @@ static void handle_iq_time(jconn conn, char *from, const char *id,
   g_free(buf);
 }
 
-// This function borrows some code from the Gaim project
+// This function borrows some code from the Pidgin project
+static void handle_iq_time202(jconn conn, char *from, const char *id,
+                              xmlnode xmldata)
+{
+  xmlnode x;
+  xmlnode myquery;
+  char *buf, *utf8_buf;
+  time_t now_t;
+  struct tm *now;
+  char const *sign;
+  int diff = 0;
+
+  time(&now_t);
+
+  scr_LogPrint(LPRINT_LOGNORM, "Received an IQ time request from <%s>", from);
+
+  buf = g_new0(char, 512);
+
+  x = jutil_iqnew(JPACKET__RESULT, NULL);
+  xmlnode_hide(xmlnode_get_tag(x, "query"));
+  xmlnode_put_attrib(xmlnode_insert_tag(x, "time"), "xmlns", NS_XMPP_TIME);
+  xmlnode_put_attrib(x, "id", id);
+  xmlnode_put_attrib(x, "to", xmlnode_get_attrib(xmldata, "from"));
+  myquery = xmlnode_get_tag(x, "time");
+
+  now = localtime(&now_t);
+
+  if (now->tm_isdst >= 0) {
+#if defined HAVE_TM_GMTOFF
+    diff = now->tm_gmtoff;
+#elif defined HAVE_TIMEZONE
+    tzset();
+    diff = -timezone;
+#endif
+  }
+
+  if (diff < 0) {
+    sign = "-";
+    diff = -diff;
+  } else {
+    sign = "+";
+  }
+  diff /= 60;
+  snprintf(buf, 512, "%c%02d:%02d", *sign, diff / 60, diff % 60);
+  if ((utf8_buf = to_utf8(buf))) {
+    xmlnode_insert_cdata(xmlnode_insert_tag(myquery, "tzo"), utf8_buf, -1);
+    g_free(utf8_buf);
+  }
+
+  now = gmtime(&now_t);
+
+  strftime(buf, 512, "%Y-%m-%dT%TZ", now);
+  xmlnode_insert_cdata(xmlnode_insert_tag(myquery, "utc"), buf, -1);
+
+  jab_send(jc, x);
+  xmlnode_free(x);
+  g_free(buf);
+}
+
+// This function borrows some code from the Pidgin project
 static void handle_iq_get(jconn conn, char *from, xmlnode xmldata)
 {
   const char *id, *ns;
@@ -1477,6 +1643,13 @@ static void handle_iq_get(jconn conn, char *from, xmlnode xmldata)
     return;
   }
 
+  x = xmlnode_get_tag(xmldata, "time");
+  ns = xmlnode_get_attrib(x, "xmlns");
+  if (ns && !strcmp(ns, NS_XMPP_TIME)) {
+    handle_iq_time202(conn, from, id, xmldata);
+    return;
+  }
+
   x = xmlnode_get_tag(xmldata, "query");
   ns = xmlnode_get_attrib(x, "xmlns");
   if (ns && !strcmp(ns, NS_DISCO_INFO)) {
@@ -1485,6 +1658,13 @@ static void handle_iq_get(jconn conn, char *from, xmlnode xmldata)
     handle_iq_disco_items(conn, from, id, xmldata);
   } else if (ns && !strcmp(ns, NS_VERSION)) {
     handle_iq_version(conn, from, id, xmldata);
+  } else if (ns && !strcmp(ns, NS_LAST)) {
+    if (!settings_opt_get_int("iq_last_disable") &&
+        (!settings_opt_get_int("iq_last_disable_when_notavail") ||
+         jb_getstatus() != notavail))
+      handle_iq_last(conn, from, id, xmldata);
+    else
+      send_iq_not_available(conn, from, xmldata);
   } else if (ns && !strcmp(ns, NS_TIME)) {
     handle_iq_time(conn, from, id, xmldata);
   } else {
